#!/bin/bash

SupervisordPidFile=/var/run/cb/cb-supervisord.pid

# shellcheck source=/dev/null
source <(/usr/share/cb/cbconf /etc/cb/cb.conf)
PgSqlPort=${PgSqlPort:?}
ClusterMembership=${ClusterMembership:?}
ProfileSvcInitBash=${ProfileSvcInitBash:-0}
CbStatsDaemon=${CbStatsDaemon:-"false"}
RabbitMQPidFile=${RabbitMQPidFile:-/var/run/cb/rabbitmq/pid}

if [ "$ProfileSvcInitBash" -ne 0 ]; then
    printf "\n\n" >> /var/log/cb/services/cb-enterprise.timing.out
    echo "--------------------------------------" >> /var/log/cb/services/cb-enterprise.timing.out

    PS4='+ $(date "+%Y-%m-%d %H:%M:%S.%N")\011 '
    exec 3>&2 2>> /var/log/cb/services/cb-enterprise.timing.out
    set -x
fi

start_dependencies() {
    rsyslogd
    crond -s
}

stop_dependencies() {
    test -f /var/run/crond.pid && kill "$(cat /var/run/crond.pid)"
    test -f /var/run/rsyslogd.pid && kill "$(cat /var/run/rsyslogd.pid)"
}

clean_volatile_stats() {
    if [ -d /var/run/cb/stats/coreservices/volatile ]; then
        rm -rf /var/run/cb/stats/coreservices/volatile
    fi
    if [ -f /var/log/cb/datastore/cb-datastore.hprof ]; then
        mv -f /var/log/cb/datastore/cb-datastore.hprof /var/log/cb/datastore/cb-datastore.hprof.old
    fi
    if [ -f /var/log/cb/solr/cb-solr.hprof ]; then
        mv -f /var/log/cb/solr/cb-solr.hprof /var/log/cb/solr/cb-solr.hprof.old
    fi
}

remove_rabbitmq_pid() {
    rm -f "${RabbitMQPidFile}"
}

wait_for_datagrid_init() {
    for i in {1..120}; do
        /usr/share/cb/cbdatagrid cluster --ready > /dev/null 2>&1
        ready_retval=$?

        if [[ $ready_retval -eq 0 ]]; then
            break
        fi

        if ((i == 5)); then
            echo -n "Waiting for cb-datagrid to initialize..."
        elif ((i % 8 == 5)); then
            echo -n "."
        fi

        sleep 0.5
    done

    if ((i >= 5)); then
        echo
    fi

    ((i < 120))
}

wait_for_rabbitmq_init() {
    echo "$(date '+%Y-%m-%d %H:%M:%S') Starting RabbitMQ" >> /var/log/cb/services/cb-rabbitmq.log
    for i in {1..30}; do
        echo "$(date '+%Y-%m-%d %H:%M:%S') Checking RabbitMQ Status" >> /var/log/cb/services/cb-rabbitmq.log
        /usr/share/cb/cbrabbitmqctl wait "${RabbitMQPidFile}" >> /var/log/cb/services/cb-rabbitmq.log 2>&1
        wait_retval=$?

        if [[ wait_retval -eq 0 ]]; then
            break
        fi

        if ((i == 2)); then
            echo -n "Waiting for cb-rabbitmq to initialize..."
        elif ((i % 3 == 2)); then
            echo -n "."
        fi

        sleep 3
    done

    if [[ $wait_retval -ne 0 ]]; then
        echo "Unable to connect to cb-rabbitmq. Check cb-rabbitmq service startup log for more details."
        echo "$(date '+%Y-%m-%d %H:%M:%S') RabbitMQ Failed to start, removing pid" >> /var/log/cb/services/cb-rabbitmq.log
        remove_rabbitmq_pid
    fi

    echo "$(date '+%Y-%m-%d %H:%M:%S') Done waiting for RabbitMQ to start" >> /var/log/cb/services/cb-rabbitmq.log
    return $wait_retval
}

wait_for_pgsql_init() {
    for i in {1..60}; do
        if psql -d cb -p "$PgSqlPort" -c '\dt' > /dev/null 2>&1; then
            break
        fi

        if ((i == 5)); then
            echo -n "Waiting for cb-pgsql to initialize..."
        elif ((i % 8 == 5)); then
            echo -n "."
        fi

        sleep 0.5
    done

    if ((i >= 5)); then
        echo
    fi

    ((i < 60))
}

wait_for_solr_init() {
    echo "   Waiting for cb-solr to build the terms dictionary."
    echo -n "   Depending on index size this may take a while..."

    for i in {1..480}; do
        status=$(/usr/share/cb/virtualenv/bin/python -m cb.maintenance.cbquery.main --solr-up 2> /dev/null)
        if ((status == 1)); then
            break
        fi

        if ((i % 5 == 0)); then
            echo -n "."
        fi

        sleep 0.5
    done

    if ((i >= 1)); then
        echo
    fi

    ((i < 480))
}

wait_for_gunicorn_init() {
    service_name=$1
    for i in {1..30}; do
        if /usr/share/cb/cbcheck gunicorn "--$service_name" > /dev/null 2>&1; then
            break
        fi

        # Wait to spew stuff to screen until this is really taking some time
        if ((i == 5)); then
            echo -n "   Waiting for cb-$service_name to initialize.."
        fi

        if ((i % 5 == 0)); then
            echo -n "."
        fi

        sleep 0.5
    done

    if ((i > 10)); then
        echo
    fi

    if ((i == 30)); then
        echo "Unable to connect to cb-$service_name API. Check $service_name startup log for more details."
    fi

    ((i < 30))
}

log_progress() {
    logger -p user.info -t "cb-services[$$]" -- "$1"
}

stop_cron_jobs() {
    count=0
    count_by=1
    timeout=${CbShutdownKillJobsTimeout:-30}

    pgrep_jobs="pgrep -f '/usr/share/cb/virtualenv/bin/python -m cb\.maintenance\.job_runner'"
    if [[ -n $(eval "${pgrep_jobs}") ]]; then

        echo "Waiting up to ${timeout} seconds for cb cronjobs to exit..."

        until [[ -z $(eval "${pgrep_jobs}") ]] || [[ ${count} -gt ${timeout} ]]; do
            sleep ${count_by}
            ((count += count_by)) || true
        done

        mapfile -t running_pids < <(eval "${pgrep_jobs}")
        if [[ ${#running_pids[@]} -ne 0 ]]; then
            echo "Killing remaining cb cronjobs."
            for pid in "${running_pids[@]}"; do
                kill -9 "${pid}"
            done
        fi
    fi
}

stop() {
    log_progress "---- Stopping cb-enterprise services.... ----"
    touch /var/lock/subsys/cb-enterprise.stopping.lock

    stop_cron_jobs

    if [ "${CbStatsDaemon,,}" = "true" ]; then
        /usr/share/cb/cbservice cb-stats stop
    fi

    /usr/share/cb/cbservice cb-nginx stop
    /usr/share/cb/cbservice cb-enterprised stop
    /usr/share/cb/cbservice cb-allianceclient stop

    if [ "$ClusterMembership" != "Minion" ]; then
        /usr/share/cb/cbservice cb-liveresponse stop
    fi

    /usr/share/cb/cbservice cb-coreservices stop
    /usr/share/cb/cbservice cb-sensorservices stop
    /usr/share/cb/cbservice cb-datastore stop
    /usr/share/cb/cbservice cb-solr stop
    /usr/share/cb/cbservice cb-rabbitmq stop
    /usr/share/cb/cbservice cb-redis stop

    /usr/share/cb/cbservice cb-datagrid stop

    if [ "$ClusterMembership" != "Minion" ]; then
        /usr/share/cb/cbservice cb-pgsql stop
    fi

    if [ -f /var/run/cb/cb-supervisord.pid ]; then
        kill "$(cat /var/run/cb/cb-supervisord.pid)"
    fi

    rm -f /var/lock/subsys/cb-enterprise.stopping.lock

    stop_dependencies

    log_progress "---- cb-enterpise services stopped ----"
}

on_subservice_start_err() {
    if ! grep '^DontStopOtherServicesOnCBEnterpriseFail=True$' /etc/cb/cb.conf > /dev/null 2>&1; then
        stop
    fi
    exit 1
}

start() {
    log_progress "---- Starting cb-enterprise($(cat /usr/share/cb/VERSION)) services.... ----"

    # Clean up orphaned lock file(s)
    rm -f /var/lock/subsys/cb-enterprise.stopping.lock

    ulimit -n "${CbFileDescriptorLimit:-80000}"

    clean_volatile_stats || exit 1

    start_dependencies || exit 1

    /usr/share/cb/virtualenv/bin/python -m cb.maintenance.cbstartup.main --stage startup
    retval=$?

    if [[ $retval -eq 2 ]]; then
        echo "VMware Carbon Black EDR services are already started."
        exit 0
    elif [[ $retval -ne 0 ]]; then
        exit $retval
    fi

    /usr/share/cb/virtualenv/bin/python -m cb.utils.cb_supervisor "/etc/cb/cb.conf"
    /usr/share/cb/supervisord -c /etc/cb/cb-supervisord.conf --pidfile "$SupervisordPidFile" || on_subservice_start_err

    if [ "$ClusterMembership" != "Minion" ]; then
        /usr/share/cb/cbservice cb-pgsql start
        wait_for_pgsql_init || on_subservice_start_err
    fi

    /usr/share/cb/cbservice cb-datagrid start
    wait_for_datagrid_init || on_subservice_start_err

    /usr/share/cb/cbservice cb-redis start || on_subservice_start_err

    remove_rabbitmq_pid
    /usr/share/cb/cbservice cb-rabbitmq start
    wait_for_rabbitmq_init || on_subservice_start_err

    /usr/share/cb/virtualenv/bin/python -m cb.maintenance.cbstartup.main --stage post-storage || on_subservice_start_err

    /usr/share/cb/cbservice cb-solr start
    wait_for_solr_init || on_subservice_start_err

    /usr/share/cb/cbservice cb-coreservices start || on_subservice_start_err
    wait_for_gunicorn_init 'coreservices' || on_subservice_start_err

    /usr/share/cb/cbservice cb-sensorservices start
    wait_for_gunicorn_init 'sensorservices' || on_subservice_start_err
    /usr/share/cb/cbservice cb-datastore start

    if [ "$ClusterMembership" != "Minion" ]; then
        /usr/share/cb/cbservice cb-liveresponse start
        wait_for_gunicorn_init 'liveresponse' || on_subservice_start_err
    fi

    /usr/share/cb/cbservice cb-allianceclient start || on_subservice_start_err
    /usr/share/cb/cbservice cb-enterprised start || on_subservice_start_err
    /usr/share/cb/cbservice cb-nginx start || on_subservice_start_err

    if [ "${CbStatsDaemon,,}" = "true" ]; then
        /usr/share/cb/cbservice cb-stats start || on_subservice_start_err
    fi

    /usr/share/cb/virtualenv/bin/python -m cb.maintenance.cbstartup.main --stage post-start || on_subservice_start_err

    log_progress "---- cb-enterprise successfully started ----"
}

status() {
    retval=0

    if ! (test -f /var/run/cb/cb-supervisord.pid && ps "$(cat /var/run/cb/cb-supervisord.pid)" > /dev/null 2>&1); then
        echo "cb-enterprise is stopped."
        exit 1
    fi

    if [ "$ClusterMembership" != "Minion" ]; then
        /usr/share/cb/cbservice cb-pgsql status || retval=1
    fi

    /usr/share/cb/cbservice cb-datagrid status || retval=1
    /usr/share/cb/cbservice cb-redis status || retval=1
    /usr/share/cb/cbservice cb-rabbitmq status || retval=1
    /usr/share/cb/cbservice cb-solr status || retval=1
    /usr/share/cb/cbservice cb-nginx status || retval=1
    /usr/share/cb/cbservice cb-datastore status || retval=1
    /usr/share/cb/cbservice cb-coreservices status || retval=1
    /usr/share/cb/cbservice cb-sensorservices status || retval=1

    if [ "$ClusterMembership" != "Minion" ]; then
        /usr/share/cb/cbservice cb-liveresponse status || retval=1
    fi

    /usr/share/cb/cbservice cb-allianceclient status || retval=1
    /usr/share/cb/cbservice cb-enterprised status || retval=1

    if [ "${CbStatsDaemon,,}" = "true" ]; then
        /usr/share/cb/cbservice cb-stats status || retval=1
    fi

    exit $retval
}

case "$1" in
    start)
        $1
        ;;
    stop)
        $1
        ;;
    status)
        $1
        ;;
esac

if [ "$ProfileSvcInitBash" -ne 0 ]; then
    set +x
    exec 2>&3 3>&-
fi

exit $?
